extern crate proc_macro2;
extern crate quote;
extern crate syn;

extern crate proc_macro;

use parse::{Parse, ParseStream};
use proc_macro::{Span, TokenStream}; use punctuated::Punctuated;
// 用于处理宏输入
use quote::quote; // 用于生成代码
use syn::*;

struct Args {
    vars: Vec<syn::Expr>
}

impl Parse for Args {
    fn parse(input: ParseStream) -> syn::parse::Result<Self> {
        let vars = Punctuated::<syn::Expr, syn::Token![,]>::parse_terminated(input)?;
        
        Ok(Args {
            vars: vars.into_iter().collect(),
        })
    }
}

impl Args {
    pub fn get_arg(&self, index: usize) -> syn::Result<syn::Expr> {
        match self.vars.get(index) {
            Some(var) => Ok(var.clone()),
            None => return Err(syn::Error::new(
                Span::call_site().into(),
                "No Route was provided"
            ))
        }
    }
}

#[allow(dead_code)]
pub fn gen_route(attr: TokenStream, item: TokenStream) -> TokenStream {
    let args = parse_macro_input!(attr as Args);
    let func = parse_macro_input!(item as ItemFn);
    
    let vis = func.vis.clone();
    let ident = func.sig.ident.clone();
    
    let method = args.get_arg(0).unwrap();
    let route = args.get_arg(1).unwrap();

    let expanded = quote! {
        #[allow(non_camel_case_types)]
        #vis struct #ident;
        
        impl #ident {
            #vis fn route() -> axum::Router {
                #func
                
                axum::Router::new().route(#route, #method (#ident))
            }
        }
    };
    
    expanded.into()
}

pub fn gen_get_route(attr: TokenStream, item: TokenStream) -> TokenStream {
    let args = parse_macro_input!(attr as Args);
    let func = parse_macro_input!(item as ItemFn);
    
    let vis = func.vis.clone();
    let ident = func.sig.ident.clone();
    
    let route = args.get_arg(0).unwrap();
    
    let expanded = quote! {
	    #vis fn #ident () -> (&'static str, axum::routing::method_routing::MethodRouter) {
		    #func
			
            (#route, axum::routing::get(#ident))
		}
	};
    expanded.into()
}

pub fn gen_post_route(attr: TokenStream, item: TokenStream) -> TokenStream {
    let args = parse_macro_input!(attr as Args);
    let func = parse_macro_input!(item as ItemFn);
    
    let vis = func.vis.clone();
    let ident = func.sig.ident.clone();
    
    let route = args.get_arg(0).unwrap();
    
    let expanded = quote! {
	    #vis fn #ident () -> (&'static str, axum::routing::method_routing::MethodRouter) {
		    #func
			
            (#route, axum::routing::post(#ident))
		}
	};
    expanded.into()
}

pub fn gen_delete_route(attr: TokenStream, item: TokenStream) -> TokenStream {
    let args = parse_macro_input!(attr as Args);
    let func = parse_macro_input!(item as ItemFn);
    
    let vis = func.vis.clone();
    let ident = func.sig.ident.clone();
    
    let route = args.get_arg(0).unwrap();
    
    let expanded = quote! {
	    #vis fn #ident () -> (&'static str, axum::routing::method_routing::MethodRouter) {
		    #func
			
            (#route, axum::routing::delete(#ident))
		}
	};
    expanded.into()
}

pub fn gen_put_route(attr: TokenStream, item: TokenStream) -> TokenStream {
    let args = parse_macro_input!(attr as Args);
    let func = parse_macro_input!(item as ItemFn);
    
    let vis = func.vis.clone();
    let ident = func.sig.ident.clone();
    
    let route = args.get_arg(0).unwrap();
    
    let expanded = quote! {
	    #vis fn #ident () -> (&'static str, axum::routing::method_routing::MethodRouter) {
		    #func
			
            (#route, axum::routing::put(#ident))
		}
	};
    expanded.into()
}

pub fn gen_option_route(attr: TokenStream, item: TokenStream) -> TokenStream {
    let args = parse_macro_input!(attr as Args);
    let func = parse_macro_input!(item as ItemFn);
    
    let vis = func.vis.clone();
    let ident = func.sig.ident.clone();
    
    let route = args.get_arg(0).unwrap();
    
    let expanded = quote! {
	    #vis fn #ident () -> (&'static str, axum::routing::method_routing::MethodRouter) {
		    #func
			
            (#route, axum::routing::option(#ident))
		}
	};
    expanded.into()
}
